package iot.system.iotserver.runners;

import iot.system.iotserver.config.CmdLineArgs;
import iot.system.iotserver.config.RuntimeConfiguration;
import iot.system.iotserver.enums.BoardStatus;
import iot.system.iotserver.enums.ClusterConnectionState;
import iot.system.iotserver.message.node.HeartbeatMessage;
import iot.system.iotserver.message.server.*;
import iot.system.iotserver.model.Board;
import iot.system.iotserver.repository.BoardRepository;
import iot.system.iotserver.support.CustomLogger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.CommandLineRunner;
import org.springframework.boot.logging.LogLevel;
import org.springframework.scheduling.annotation.Async;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;

import java.time.Instant;
import java.util.Hashtable;
import java.util.Map;
import java.util.Random;

@Component
public class Main implements CommandLineRunner {

    @Autowired
    private BoardRepository boardRepository;

    @Override
    @Async
    public void run(String... args) throws Exception {
        CustomLogger.log("Sending request to join cluster");
        //Send cluster connection request
        ConnectionMessage msg = new ConnectionMessage();
        msg.setTargetID(0);
        msg.setSenderID(CmdLineArgs.getServerID());
        msg.setConnect(true);
        msg.setGroupID(CmdLineArgs.getConsumerGroup());
        msg.setTimestamp(Instant.now());
        MessageProcessor.getInstance().sendMessage("S2S", msg);

        CustomLogger.log("Waiting for response");
        //Wait for result
        Long startTime = Instant.now().toEpochMilli();
        while((Instant.now().toEpochMilli()-startTime<10000)&&(RuntimeConfiguration.getInstance().getClusterConnectionState()==ClusterConnectionState.NOT_CONNECTED)){
            try{
                Thread.sleep(200);
            }
            catch (InterruptedException e){
                //We can just go again to wait
            }
        }

        if(RuntimeConfiguration.getInstance().getClusterConnectionState()==ClusterConnectionState.NOT_CONNECTED){
            CustomLogger.log("No response from cluster. Assuming leadership as one and only node in cluster");
            RuntimeConfiguration.getInstance().setLeader(CmdLineArgs.getServerID());
            RuntimeConfiguration.getInstance().setClusterConnectionState(ClusterConnectionState.CONNECTED);
            RuntimeConfiguration.getInstance().getTopology().put(CmdLineArgs.getServerID(),CmdLineArgs.getServerID());
            CustomLogger.log("Topology: "+RuntimeConfiguration.getInstance().topologyToString());
        }
        else if(RuntimeConfiguration.getInstance().getClusterConnectionState()==ClusterConnectionState.DENIED){
            CustomLogger.log("Connection denied");
        }
        else if(RuntimeConfiguration.getInstance().getClusterConnectionState()==ClusterConnectionState.CONNECTED){
            CustomLogger.log("Connected");
        }
    }

    @Scheduled(fixedRate = 500)
    public void serverHeartbeat(){
        if(RuntimeConfiguration.getInstance().getClusterConnectionState()!=ClusterConnectionState.CONNECTED) return;

        ServerHeartbeatMessage message = new ServerHeartbeatMessage();
        message.setTimestamp(Instant.now());
        message.setSenderID(CmdLineArgs.getServerID());
        message.setTargetID(0);
        MessageProcessor.getInstance().sendMessage("S2S",message);
    }

    @Scheduled(fixedRate = 1000)
    public void clusterHeartbeatCheck(){
        if(RuntimeConfiguration.getInstance().getClusterConnectionState()!=ClusterConnectionState.CONNECTED) return;

        //Map<Integer, Instant> clusterActivity = new Hashtable<>(RuntimeConfiguration.getInstance().getClusterActivity());
        Map<Integer, Instant> clusterActivity = RuntimeConfiguration.getInstance().getClusterActivity();
        if(RuntimeConfiguration.getInstance().isLeader()){
            for(Integer serverID: clusterActivity.keySet()){
                Instant lastActive = clusterActivity.get(serverID);
                if((Instant.now().toEpochMilli()-lastActive.toEpochMilli()>2000)){
                    //Update and send topology
                    Map<Integer, Integer> topology = RuntimeConfiguration.getInstance().getTopology();
                    Main.removeServerFromTopology(topology,serverID);

                    TopologyUpdateMessage tum = new TopologyUpdateMessage();
                    tum.setSenderID(CmdLineArgs.getServerID());
                    tum.setTargetID(0);
                    tum.setTimestamp(Instant.now());
                    tum.setTopology(topology);
                    MessageProcessor.getInstance().sendMessage("S2S", tum);

                    //remove from table
                    clusterActivity.remove(serverID);
                    CustomLogger.log("Server with ID: "+serverID+" disconnected from cluster", LogLevel.WARN);
                }
            }
        }
        else{
            Instant leaderLastActivity = RuntimeConfiguration.getInstance().getClusterActivity().get(RuntimeConfiguration.getInstance().getLeader());
            if((Instant.now().toEpochMilli()-leaderLastActivity.toEpochMilli()>1000)){
                CustomLogger.log("Leader lost!", LogLevel.WARN);
                //Local temporary topology fix
                Map<Integer, Integer> topology = RuntimeConfiguration.getInstance().getTopology();
                Integer leaderId = RuntimeConfiguration.getInstance().getLeader();
                Main.removeServerFromTopology(topology,leaderId);
                CustomLogger.log("Temporary topology fixed calculated.", LogLevel.WARN);

                if(topology.size()==1){
                    CustomLogger.log("Last server standing! Assuming leadership.", LogLevel.WARN);
                    RuntimeConfiguration.getInstance().setLeader(CmdLineArgs.getServerID());
                    return;
                }


                CustomLogger.log("Starting election.", LogLevel.INFO);
                //Start leader election
                LeaderElectionMessage message = new LeaderElectionMessage();
                message.setSenderID(CmdLineArgs.getServerID());
                message.setTargetID(RuntimeConfiguration.getInstance().getTopology().get(CmdLineArgs.getServerID()));
                message.setTimestamp(Instant.now());
                message.setLeaderCandidate(CmdLineArgs.getServerID());
                MessageProcessor.getInstance().sendMessage("S2S",message);
                RuntimeConfiguration.getInstance().setElectionRunning(true);
                RuntimeConfiguration.getInstance().setElectionParticipant(true);
            }
        }
    }

    private static void removeServerFromTopology(Map<Integer, Integer> topology, Integer serverToRemove){
        Integer serverNext = topology.remove(serverToRemove);
        for(Integer serverID: topology.keySet()){
            Integer next = topology.get(serverID);
            if(serverToRemove.equals(next)){
                topology.put(serverID,serverNext);
                break;
            }
        }
    }

    @Scheduled(fixedRate = 1000)
    @Transactional
    public void leaderBoardStatusCheckAndUpdate(){
        if(!(RuntimeConfiguration.getInstance().isLeader()&&(RuntimeConfiguration.getInstance().getClusterConnectionState()==ClusterConnectionState.CONNECTED))){
            return;
        }

        for(Board board: boardRepository.findAll()){
            Instant lastActive = board.getLastActive();
            Integer commInterval = board.getCommInterval();
            Integer deadMultiplier = board.getDeadMultiplier();

            Long deadTimeout = lastActive.toEpochMilli() + (commInterval * deadMultiplier);

            if((deadTimeout<Instant.now().toEpochMilli())&&(board.getStatus()!=BoardStatus.OFFLINE)){
                board.setStatus(BoardStatus.OFFLINE);
                CustomLogger.log("Board [login: "+board.getLogin()+" / name: "+board.getName()+"] went OFFLINE!", LogLevel.WARN);
            }
            else if((deadTimeout>Instant.now().toEpochMilli())&&(board.getStatus()!=BoardStatus.ONLINE)){
                board.setStatus(BoardStatus.ONLINE);
                CustomLogger.log("Board [login: "+board.getLogin()+" / name: "+board.getName()+"] is now ONLINE!", LogLevel.INFO);
            }
        }
    }
}
