package iot.system.iotserver.runners;

import iot.system.iotserver.config.CmdLineArgs;
import iot.system.iotserver.config.RuntimeConfiguration;
import iot.system.iotserver.enums.ClusterConnectionState;
import iot.system.iotserver.enums.InterfaceSettingsParams;
import iot.system.iotserver.message.node.BoardConfigMessage;
import iot.system.iotserver.message.ExchangeMessage;
import iot.system.iotserver.message.node.HeartbeatMessage;
import iot.system.iotserver.message.node.InterfaceConfigMessage;
import iot.system.iotserver.message.node.InterfaceStateMessage;
import iot.system.iotserver.message.server.*;
import iot.system.iotserver.model.*;
import iot.system.iotserver.modules.AbstractModule;
import iot.system.iotserver.repository.BoardRepository;
import iot.system.iotserver.repository.InterfaceRepository;
import iot.system.iotserver.repository.ModuleVarRepository;
import iot.system.iotserver.support.CustomLogger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.logging.LogLevel;
import org.springframework.kafka.core.KafkaTemplate;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;

import javax.annotation.PostConstruct;
import javax.persistence.OneToMany;
import java.time.Instant;
import java.util.*;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;

@Component
public class MessageProcessor {

    @Autowired
    private InterfaceRepository interfaceRepository;

    @Autowired
    private ModuleVarRepository moduleVarRepository;

    @Autowired
    private BoardRepository boardRepository;

    @Autowired
    private KafkaTemplate<String, ExchangeMessage> kafkaProducer;

    private static MessageProcessor instance;

    @PostConstruct
    private void init(){
        MessageProcessor.instance = this;
    }

    @Transactional
    public void processMessage(InterfaceStateMessage message){
        List<Interface> interfaces = interfaceRepository.findInterfacesByBoard_LoginAndUidIn(message.getNodeLogin(),message.getInterfacesStates().keySet());
        List<ModuleVar> moduleVars = moduleVarRepository.findModuleVarsByAttachedInterfaceIn(interfaces);
        Set<Module> modules = new HashSet<>();
        for(ModuleVar modVar: moduleVars){
            modules.add(modVar.getModule());
        }
        for(Module mod: modules){
            AbstractModule module = AbstractModule.createInstanceFromModule(mod);
            module.processMessage(message);
        }
    }

    @Transactional
    public void processMessage(BoardConfigMessage message){
        if(!message.isConfigRequest()) return;

        Board board = boardRepository.findBoardByLogin(message.getNodeLogin());

        BoardConfigMessage response = new BoardConfigMessage();
        response.setTimestamp(Instant.now());
        response.setBoardName(board.getName());
        response.setNodeLogin(board.getLogin());
        response.setConfigRequest(false);
        response.setCommInterval(board.getCommInterval());

        InterfaceConfigMessage interfaceConfig = new InterfaceConfigMessage();
        Map<String, Map<InterfaceSettingsParams, String>> payload = new HashMap<>();
        for(Interface intf: board.getInterfaces()){
            Map<InterfaceSettingsParams, String> intfCfgParams = new HashMap<>();
            for(InterfaceConfig cfg: intf.getConfig()){
                intfCfgParams.put(cfg.getParameter(),cfg.getValue());
            }
            payload.put(intf.getUid(),intfCfgParams);
        }
        interfaceConfig.setInterfacesConfig(payload);

        response.setInterfacesConfig(interfaceConfig);

        this.sendMessage("S2N", response);
    }

    @Transactional
    public void processMessage(HeartbeatMessage message){
        Board board = boardRepository.findBoardByLogin(message.getNodeLogin());
        board.setLastActive(message.getTimestamp());
    }

    public void processMessage(ConnectionMessage message){
        if(!RuntimeConfiguration.getInstance().isLeader()) return;
        CustomLogger.log("New server with ID "+message.getSenderID()+" is requesting to join cluster", LogLevel.INFO);

        if((!CmdLineArgs.getConsumerGroup().equals(message.getGroupID()))||(RuntimeConfiguration.getInstance().getTopology().containsKey(message.getSenderID()))){
            ConnectionDeniedMessage rejectMessage = new ConnectionDeniedMessage();
            rejectMessage.setSenderID(CmdLineArgs.getServerID());
            rejectMessage.setTargetID(message.getSenderID());
            rejectMessage.setTimestamp(Instant.now());
            rejectMessage.setRejectedNodeID(message.getSenderID());
            this.sendMessage("S2S", rejectMessage);
            CustomLogger.log("New server with ID "+message.getSenderID()+" REJECTED. ID already exits in cluster OR joining in another consumer group.", LogLevel.INFO);
            return;
        }

        ConnectionAcceptedMessage response = new ConnectionAcceptedMessage();
        response.setTimestamp(Instant.now());
        response.setTargetID(message.getSenderID());
        response.setSenderID(CmdLineArgs.getServerID());
        this.sendMessage("S2S",response);
        CustomLogger.log("New server with ID "+message.getSenderID()+" ACCEPTED", LogLevel.INFO);

        Map<Integer,Integer> topology = RuntimeConfiguration.getInstance().getTopology();
        Integer leaderID = CmdLineArgs.getServerID();
        Integer newNode = response.getTargetID();
        Integer nextNode = topology.put(leaderID,newNode);
        topology.put(newNode,nextNode);

        TopologyUpdateMessage tum = new TopologyUpdateMessage();
        tum.setSenderID(CmdLineArgs.getServerID());
        tum.setTargetID(0);
        tum.setTopology(topology);
        tum.setTimestamp(Instant.now());
        MessageProcessor.getInstance().sendMessage("S2S", tum);
        CustomLogger.log("New topology sent: "+RuntimeConfiguration.getInstance().topologyToString());
    }

    public void processMessage(ConnectionDeniedMessage message){
        if(message.getRejectedNodeID()==CmdLineArgs.getServerID()){
            RuntimeConfiguration.getInstance().setClusterConnectionState(ClusterConnectionState.DENIED);
        }
    }

    public void processMessage(ConnectionAcceptedMessage message){
        RuntimeConfiguration.getInstance().setLeader(message.getSenderID());
        RuntimeConfiguration.getInstance().setClusterConnectionState(ClusterConnectionState.CONNECTED);
    }

    public void processMessage(TopologyUpdateMessage message){
        RuntimeConfiguration.getInstance().setTopology(message.getTopology());
        CustomLogger.log("New topology received: "+RuntimeConfiguration.getInstance().topologyToString());
    }

    public void processMessage(ServerHeartbeatMessage message){
        RuntimeConfiguration.getInstance().getClusterActivity().put(message.getSenderID(),message.getTimestamp());
    }

    public void processMessage(LeaderElectionMessage message){
        if(message.getLeaderCandidate()>CmdLineArgs.getServerID()){
            RuntimeConfiguration.getInstance().setElectionParticipant(true);
            Integer nextHop = RuntimeConfiguration.getInstance().getTopology().get(CmdLineArgs.getServerID());
            message.setSenderID(CmdLineArgs.getServerID());
            message.setTargetID(nextHop);
            message.setTimestamp(Instant.now());
            this.sendMessage("S2S",message);
            CustomLogger.log("Leader election message forwarded unchanged to server with ID "+nextHop+". Leader candidate ID "+message.getLeaderCandidate(), LogLevel.INFO);
        }
        if(message.getLeaderCandidate()<CmdLineArgs.getServerID()){
            if(!RuntimeConfiguration.getInstance().isElectionParticipant()){
                RuntimeConfiguration.getInstance().setElectionParticipant(true);
                Integer nextHop = RuntimeConfiguration.getInstance().getTopology().get(CmdLineArgs.getServerID());
                message.setSenderID(CmdLineArgs.getServerID());
                message.setTargetID(nextHop);
                message.setLeaderCandidate(CmdLineArgs.getServerID());
                message.setTimestamp(Instant.now());
                this.sendMessage("S2S",message);
                CustomLogger.log("Leader election message forwarded to server with ID "+nextHop+". Leader candidate ID "+message.getLeaderCandidate(), LogLevel.INFO);
            }
            else{
                CustomLogger.log("Leader election message thrown away.", LogLevel.INFO);
                return;
            }
        }
        if(message.getLeaderCandidate().equals(CmdLineArgs.getServerID())){
            CustomLogger.log("This server was elected as leader!", LogLevel.INFO);
            RuntimeConfiguration.getInstance().setLeader(CmdLineArgs.getServerID());

            LeaderElectedMessage resultMessage = new LeaderElectedMessage();
            resultMessage.setSenderID(CmdLineArgs.getServerID());
            resultMessage.setTargetID(0);
            resultMessage.setTimestamp(Instant.now());
            resultMessage.setElectedLeader(CmdLineArgs.getServerID());
            this.sendMessage("S2S",resultMessage);
            CustomLogger.log("Leader Elected message sent to all servers", LogLevel.INFO);

            RuntimeConfiguration.getInstance().setElectionRunning(false);
            RuntimeConfiguration.getInstance().setElectionParticipant(false);
        }
    }

    public void processMessage(LeaderElectedMessage message){
        CustomLogger.log("Leader Elected message received. New leader is server with ID "+message.getElectedLeader(), LogLevel.INFO);
        RuntimeConfiguration.getInstance().setElectionRunning(false);
        RuntimeConfiguration.getInstance().setElectionParticipant(false);
        RuntimeConfiguration.getInstance().setLeader(message.getElectedLeader());
    }

    public void sendMessage(String topic, ExchangeMessage message){
        kafkaProducer.send(topic,message);
    }

    static public MessageProcessor getInstance(){
        return MessageProcessor.instance;
    }

}
