/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package board.utils;

import board.io.Interface;
import board.io.InterfaceDirection;
import board.io.InterfaceSettingsParams;
import board.io.InterfaceType;
import board.io.uart.UartListener;
import board.io.uart.UartPacketListener;
import com.fazecast.jSerialComm.SerialPort;

import java.io.FileWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author janzubr
 */
public class AllwinnerUtils extends BoardUtils{
    private static final String GPIO_PATH = "/sys/class/gpio";
    private static final String EXPORT_PATH = GPIO_PATH+"/export";
    private static final String UNEXPORT_PATH = GPIO_PATH+"/unexport";

    private final ConcurrentHashMap<Integer,FileWriter> outPinCache;
    private final ConcurrentHashMap<String, SerialPort> serialPorts;
    private final ConcurrentHashMap<String, SerialPort> openSerialPorts;
    
    protected AllwinnerUtils(){
        super();
        this.outPinCache = new ConcurrentHashMap<>();
        this.serialPorts = new ConcurrentHashMap<>();
        this.openSerialPorts = new ConcurrentHashMap<>();

        SerialPort[] ports = SerialPort.getCommPorts();
        for(SerialPort port:ports){
            String portName = port.getSystemPortName();
            serialPorts.put(portName,port);
        }
    }

    private int getPinNum(Interface intf){
        if(intf.getDescriptor().getType() != InterfaceType.PIN){
            return -1;
        }

        int[] UID = intf.getDescriptor().getIntArrayOfUID();
        int portNum = UID[4];
        int pinNum = UID[5];

        return portNum*32+pinNum;
    }

    @Override
    protected boolean openPin(Interface intf) {
        int pinNum = this.getPinNum(intf);
        intf.setDirection(InterfaceDirection.fromString(intf.getSettingsParameters().get(InterfaceSettingsParams.GPIO_DIR)));

        String pinDir = "out";
        if(intf.getDirection()==InterfaceDirection.IN){
            pinDir = "in";
        }

        try{
            FileWriter fileWriter = new FileWriter(EXPORT_PATH);
            fileWriter.write(Integer.toString(pinNum));
            fileWriter.close();

            fileWriter = new FileWriter(GPIO_PATH+"/gpio"+pinNum+"/direction");
            fileWriter.write(pinDir);
            fileWriter.close();

            this.readPin(intf);
            return true;
        }
        catch(IOException e){
            return false;
        }
    }

    @Override
    protected boolean openUart(Interface intf) {
        int[] UID = intf.getDescriptor().getIntArrayOfUID();

        String portName = "";
        if(UID[4]==0){
            portName = "ttyS";
        }
        else if(UID[4]==1){
            portName = "ttyUSB";
        }
        portName = portName + Integer.toString(UID[5]);

        if(!this.serialPorts.containsKey(portName)) return false;
        if(this.openSerialPorts.containsKey(intf.getDescriptor().getUid())) return true;

        try{
            Map<InterfaceSettingsParams,String> settings = intf.getSettingsParameters();
            SerialPort port = this.serialPorts.get(portName);

            port.setBaudRate(Integer.parseInt(settings.get(InterfaceSettingsParams.UART_BAUD_RATE)));
            port.setParity(Integer.parseInt(settings.get(InterfaceSettingsParams.UART_PARITY)));
            port.setFlowControl(Integer.parseInt(settings.get(InterfaceSettingsParams.UART_FLOW_CONTROL)));
            port.setNumStopBits(Integer.parseInt(settings.get(InterfaceSettingsParams.UART_STOP_BITS)));

            if(settings.containsKey(InterfaceSettingsParams.UART_PACKET_SIZE)){
                port.addDataListener(new UartPacketListener(intf,Integer.parseInt(settings.get(InterfaceSettingsParams.UART_PACKET_SIZE))));
            }
            else{
                port.addDataListener(new UartListener(intf));
            }

            if(port.openPort()){
                this.openSerialPorts.put(intf.getDescriptor().getUid(),port);
                return true;
            }
            else{
                return false;
            }
        }
        catch(Exception e){
            System.err.println(e);
            return false;
        }
    }

    @Override
    protected boolean writePin(Interface intf) {
        int pinNum = this.getPinNum(intf);

        if(intf.getDataOut().length != 1){
            return false;
        }

        String val = "0";
        if(intf.getDataOut()[0] == 0x01){
            val = "1";
        }

        try {
            FileWriter fileWriter;
            if(this.outPinCache.containsKey(pinNum)){
                fileWriter = this.outPinCache.get(pinNum);
            }
            else{
                fileWriter = new FileWriter(GPIO_PATH+"/gpio"+pinNum+"/value");
                this.outPinCache.put(pinNum, fileWriter);
            }
            fileWriter.write(val);
            fileWriter.flush();

            intf.setDataOut(new byte[]{});
            return true;
        } catch (IOException ex) {
            Logger.getLogger(AllwinnerUtils.class.getName()).log(Level.SEVERE, null, ex);
            return false;
        }
    }

    @Override
    protected boolean writeUart(Interface intf) {
        if(!this.openSerialPorts.containsKey(intf.getDescriptor().getUid())) return false;

        byte[] bytesToWrite = intf.getDataOutBytes();
        int writtenBytes = this.openSerialPorts.get(intf.getDescriptor().getUid()).writeBytes(bytesToWrite,bytesToWrite.length);
        return writtenBytes > -1;
    }

    @Override
    protected boolean readPin(Interface intf) {
        int pinNum = this.getPinNum(intf);

        try {
            String content = Files.readAllLines(Paths.get(GPIO_PATH+"/gpio"+pinNum+"/value")).get(0);
            if(content.equals("0")){
                intf.setDataIn(new int[]{0x00});
            }
            else if(content.equals("1")){
                intf.setDataIn(new int[]{0x01});
            }
            else{
                throw new RuntimeException("This shit should not happen");
            }
            return true;
        } catch (IOException ex) {
            return false;
        }
    }

    @Override
    protected boolean readUart(Interface intf) {
        //UART is read by listener so interface of UART type
        //has always latest data in dataIn property
        //Just return true then
        return true;
    }

    @Override
    protected boolean closePin(Interface intf) {
        int pinNum = this.getPinNum(intf);

        try{
            this.outPinCache.remove(pinNum);
            FileWriter fileWriter = new FileWriter(UNEXPORT_PATH);
            fileWriter.write(Integer.toString(pinNum));
            fileWriter.close();
            return true;
        }
        catch(IOException e){
            return false;
        }
    }

    @Override
    protected boolean closeUart(Interface intf) {
        if(!this.openSerialPorts.containsKey(intf.getDescriptor().getUid())) return true;

        SerialPort port = this.openSerialPorts.get(intf.getDescriptor().getUid());

        if(port.closePort()){
            this.openSerialPorts.remove(intf.getDescriptor().getUid());
            port.removeDataListener();
            return true;
        }
        else{
            return false;
        }
    }
}
