package exchange;

import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import config.Config;
import exchange.dto.BoardConfigMessage;
import exchange.dto.ExchangeMessage;
import exchange.dto.HeartbeatMessage;
import org.apache.kafka.clients.consumer.*;
import org.apache.kafka.clients.producer.KafkaProducer;
import org.apache.kafka.clients.producer.Producer;
import org.apache.kafka.clients.producer.ProducerConfig;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.common.protocol.types.Field;
import org.apache.kafka.common.serialization.LongDeserializer;
import org.apache.kafka.common.serialization.LongSerializer;
import org.apache.kafka.common.serialization.StringDeserializer;
import org.apache.kafka.common.serialization.StringSerializer;
import support.ResourceCache;

import java.io.IOException;
import java.time.Duration;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.time.temporal.TemporalUnit;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.concurrent.*;

public class ExchangeKafka extends Exchange {
    private static String BROKERS;
    private static Integer MSG_COUNT;
    private static String CLIENT_ID;
    private static String TOPIC_NAME_OUT = "N2S";
    private static String TOPIC_NAME_IN = "S2N";
    private static String CONSUMER_GROUP_ID;
    private static Integer MAX_NUM_MSG_FOUND;
    private static String OFFSET_LATEST = "latest";
    private static String OFFSET_EARLIEST = "earliest";
    private static Integer MAX_POLL_RECORDS = 1000;

    private Producer<Long, String> producer;
    private Consumer<Long, String> consumer;
    private Thread consumerThread;

    private final BlockingQueue<ExchangeMessage> recievedMsgs;

    private final ScheduledExecutorService heartbeatService;

    protected ExchangeKafka(){
        System.out.println("Kafka exchange init START");
        ExchangeKafka.CLIENT_ID = Config.getLogin();
        ExchangeKafka.BROKERS = Config.getServerURL();
        ExchangeKafka.CONSUMER_GROUP_ID = "consumer-"+Config.getLogin();

        this.recievedMsgs = new LinkedBlockingQueue<>();

        this.producer = ExchangeKafka.createProducer();
        this.consumer = ExchangeKafka.createConsumer();

        this.consumerThread = new Thread(new ConsumerThread(this.consumer,this.recievedMsgs));
        this.consumerThread.setDaemon(true);
        this.consumerThread.start();
        System.out.println("Kafka exchange init DONE. Requesting config...");

        BoardConfigMessage cfgmsg = new BoardConfigMessage();
        cfgmsg.setConfigRequest(true);
        this.pushMessage(cfgmsg);
        System.out.println("Request send. Waiting for response...");

        boolean configLoaded = false;
        long startTime = Instant.now().toEpochMilli();
        int skippedMsgCount = 0;
        while((!configLoaded)&&(Instant.now().toEpochMilli()-startTime<10000)){
            try{
                ExchangeMessage topMsg = this.recievedMsgs.peek();
                if(topMsg == null) continue;
                if((topMsg.getTimestamp().toEpochMilli()>cfgmsg.getTimestamp().toEpochMilli())&&(topMsg instanceof BoardConfigMessage)){
                    configLoaded = true;
                }
                else{
                    this.recievedMsgs.poll(100, TimeUnit.MILLISECONDS);
                    skippedMsgCount++;
                }
            }
            catch(InterruptedException e){
                System.out.println("This is friendly exception: "+e+" We can ignore that one.");
            }
        }

        if(configLoaded){
            long runtime = Instant.now().toEpochMilli() - startTime;
            System.out.println("Config response received in "+runtime+"ms. Skipped "+skippedMsgCount+" old messages.");

            BoardConfigMessage confMsg = (BoardConfigMessage) this.recievedMsgs.peek();

            System.out.println("Starting heartbeat service. Heartbeat every "+confMsg.getCommInterval()+"ms.");
            this.heartbeatService = Executors.newSingleThreadScheduledExecutor(new ExchangeKafka.ExchangeThreadFactory());
            this.heartbeatService.scheduleAtFixedRate(new ExchangeKafka.HeartbeatThread(), 0, confMsg.getCommInterval(), TimeUnit.MILLISECONDS);
        }
        else{
            System.out.println("Config did not arrived! Shutting down");
            throw new RuntimeException(new TimeoutException("Config request timeout"));
        }

    }

    private static Producer<Long, String> createProducer(){
        Properties props = new Properties();
        props.put(ProducerConfig.BOOTSTRAP_SERVERS_CONFIG,ExchangeKafka.BROKERS);
        props.put(ProducerConfig.CLIENT_ID_CONFIG, ExchangeKafka.CLIENT_ID);
        props.put(ProducerConfig.KEY_SERIALIZER_CLASS_CONFIG, LongSerializer.class.getName());
        props.put(ProducerConfig.VALUE_SERIALIZER_CLASS_CONFIG, StringSerializer.class.getName());
        return new KafkaProducer<>(props);
    }

    private static Consumer<Long, String> createConsumer(){
        Properties props = new Properties();
        props.put(ConsumerConfig.BOOTSTRAP_SERVERS_CONFIG, ExchangeKafka.BROKERS);
        props.put(ConsumerConfig.GROUP_ID_CONFIG, ExchangeKafka.CONSUMER_GROUP_ID);
        props.put(ConsumerConfig.KEY_DESERIALIZER_CLASS_CONFIG, LongDeserializer.class.getName());
        props.put(ConsumerConfig.VALUE_DESERIALIZER_CLASS_CONFIG, StringDeserializer.class.getName());
        props.put(ConsumerConfig.MAX_POLL_RECORDS_CONFIG, ExchangeKafka.MAX_POLL_RECORDS);
        props.put(ConsumerConfig.ENABLE_AUTO_COMMIT_CONFIG, "false");
        props.put(ConsumerConfig.AUTO_OFFSET_RESET_CONFIG, ExchangeKafka.OFFSET_LATEST);
        Consumer<Long, String> consumer = new KafkaConsumer<>(props);
        consumer.subscribe(Collections.singletonList(ExchangeKafka.TOPIC_NAME_IN));
        return consumer;
    }

    @Override
    public void pushMessage(ExchangeMessage message) {
        try{
            String messageJson = ResourceCache.getInstance().getObjectMapper().writeValueAsString(message);
            ProducerRecord<Long, String> record = new ProducerRecord<>(ExchangeKafka.TOPIC_NAME_OUT,messageJson);
            this.producer.send(record);
        }
        catch(Exception e){
        }
    }

    @Override
    public ExchangeMessage fetchMessage() {
        try{
            return this.recievedMsgs.take();
        }
        catch(InterruptedException ex){
            return null;
        }
    }

    private class ConsumerThread implements Runnable{
        private Consumer<Long, String> consumer;
        private BlockingQueue<ExchangeMessage> recievedMsgs;
        private TypeReference<ExchangeMessage> typeReference;
        private ObjectMapper OM;

        public ConsumerThread(Consumer<Long, String> consumer, BlockingQueue<ExchangeMessage> recievedMsgs) {
            this.consumer = consumer;
            this.recievedMsgs = recievedMsgs;
            this.typeReference = new TypeReference<ExchangeMessage>(){};
            this.OM = ResourceCache.getInstance().getObjectMapper();
        }

        @Override
        public void run() {
            while(true){
                ConsumerRecords<Long, String> consumerRecords = consumer.poll(Duration.of(10, ChronoUnit.MILLIS));
                if (consumerRecords.count() == 0) continue;

                Iterator<ConsumerRecord<Long, String>> it = consumerRecords.iterator();
                while(it.hasNext()){
                    ConsumerRecord<Long, String> record = it.next();
                    String jsonData = record.value();

                    try {
                        ExchangeMessage msg = OM.readValue(jsonData, this.typeReference);
                        if(Config.getLogin().equals(msg.getNodeLogin())){
                            this.recievedMsgs.add(msg);
                        }
                    } catch (IOException e) {
                        e.printStackTrace();
                    }
                    consumer.commitAsync();
                }
            }
        }
    }

    private class HeartbeatThread implements Runnable{

        @Override
        public void run() {
            pushMessage(new HeartbeatMessage());
        }
    }

    private class ExchangeThreadFactory implements ThreadFactory {

        @Override
        public Thread newThread(Runnable r) {
            Thread T = new Thread(r);
            T.setDaemon(true);
            return T;
        }

    }
}
