package exchange;
;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import config.Config;
import exchange.auth.AuthHandler;
import exchange.dto.BoardConfigMessage;
import exchange.dto.ExchangeMessage;
import org.apache.http.HttpResponse;
import org.apache.http.client.HttpClient;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.conn.HttpClientConnectionManager;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.HttpClients;
import org.apache.http.impl.conn.PoolingHttpClientConnectionManager;
import support.ResourceCache;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.time.Instant;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.*;
import java.util.logging.Level;
import java.util.logging.Logger;

public class ExchangeRest extends Exchange {
    private static final String API = "node";
    private static final String MSGS_ENDPOINT = API+"/messages";

    private final String serverURL;
    private final HttpClient httpClient;
    private final AuthHandler AH;

    private final BlockingQueue<ExchangeMessage> recievedMsgs;
    private final BlockingDeque<ExchangeMessage> transmitMsgs;

    private final ScheduledExecutorService commService;

    protected ExchangeRest(){
        this.serverURL = Config.getServerURL();
        System.out.println("Using API: "+this.serverURL);

        System.out.println("Initializing http client...");
        //this.httpClient = HttpClientBuilder.create().build();
        HttpClientConnectionManager poolingConnManager = new PoolingHttpClientConnectionManager();
        ((PoolingHttpClientConnectionManager) poolingConnManager).setMaxTotal(10);
        RequestConfig config = RequestConfig.custom().setConnectTimeout(2000).setConnectionRequestTimeout(2000).setSocketTimeout(2000).build();
        this.httpClient = HttpClients.custom().setConnectionManager(poolingConnManager).setConnectionTimeToLive(250,TimeUnit.MILLISECONDS).setDefaultRequestConfig(config).build();
        this.AH = new AuthHandler(Config.getLogin(),Config.getPassword());
        System.out.println("Done!");

        System.out.println("Loading config from server...");
        BoardConfigMessage confMsg = (BoardConfigMessage) this.getMessageFromEndpoint(API+"/config");
        Config.setFromConfigMessage(confMsg);
        System.out.println("Config loaded!");

        this.recievedMsgs = new LinkedBlockingQueue<>();
        this.recievedMsgs.add(confMsg);

        this.transmitMsgs = new LinkedBlockingDeque<>();

        this.commService = Executors.newSingleThreadScheduledExecutor(new ExchangeRest.ExchangeThreadFactory());
        this.commService.scheduleAtFixedRate(new ExchangeRest.CommunicationThread(), 0, confMsg.getCommInterval(), TimeUnit.MILLISECONDS);
    }

    private List<ExchangeMessage> getAllMessagesFromEndpoint(String endpoint){
        try{
            HttpGet request = new HttpGet(this.serverURL+"/"+endpoint);
            request.addHeader("Authorization", AH.getAuthString(endpoint));
            System.out.println("Reading messages start");
            HttpResponse response = this.httpClient.execute(request);
            request.releaseConnection();
            System.out.println("Reading messages done");

            if(response.getStatusLine().getStatusCode()!=200){
                //throw new RuntimeException("Init failed with code: "+response.getStatusLine().getStatusCode());
                return new ArrayList<>();
            }

            BufferedReader in = new BufferedReader(new InputStreamReader(response.getEntity().getContent()));
            String responseLine;
            StringBuilder sb = new StringBuilder();
            while((responseLine = in.readLine()) != null){
                sb.append(responseLine);
            }
            in.close();

            if(sb.toString().length()<1) return null;

            ObjectMapper OM = ResourceCache.getInstance().getObjectMapper();

            return OM.readValue(sb.toString(), new TypeReference<List<ExchangeMessage>>(){});
        } catch (Throwable ex) {
            Logger.getLogger(ExchangeRest.class.getName()).log(Level.SEVERE, null, ex);
            return new ArrayList<>();
        }
    }

    private ExchangeMessage getMessageFromEndpoint(String endpoint){
        try{
            HttpGet request = new HttpGet(this.serverURL+"/"+endpoint);
            request.addHeader("Authorization", AH.getAuthString(endpoint));
            HttpResponse response = this.httpClient.execute(request);

            if(response.getStatusLine().getStatusCode()!=200){
                throw new RuntimeException("Init failed with code: "+response.getStatusLine().getStatusCode());
            }

            BufferedReader in = new BufferedReader(new InputStreamReader(response.getEntity().getContent()));
            String responseLine;
            StringBuilder sb = new StringBuilder();
            while((responseLine = in.readLine()) != null){
                sb.append(responseLine);
            }
            in.close();

            if(sb.toString().length()<1) return null;

            ObjectMapper OM = ResourceCache.getInstance().getObjectMapper();

            return OM.readValue(sb.toString(), new TypeReference<ExchangeMessage>(){});
        } catch (Exception ex) {
            Logger.getLogger(ExchangeRest.class.getName()).log(Level.SEVERE, null, ex);
            return null;
        }
    }

    private int postAllMessagesToEndpoint(String endpoint){
        try{
            LinkedList<ExchangeMessage> msgs = new LinkedList<>();
            while(!this.transmitMsgs.isEmpty()){
                msgs.add(this.transmitMsgs.take());
            }

            System.out.println("Sending "+msgs.size()+" messages!");

            String jsonData = ResourceCache.getInstance().getObjectMapper().writerFor(new TypeReference<LinkedList<ExchangeMessage>>(){}).writeValueAsString(msgs);

            HttpPost request = new HttpPost(this.serverURL+"/"+endpoint);
            request.addHeader("Authorization", AH.getAuthString(endpoint));
            request.setHeader("Content-type","application/json");

            request.setEntity(new StringEntity(jsonData));
            System.out.println("Execute REST query");

            HttpResponse response = this.httpClient.execute(request);
            request.releaseConnection();
            System.out.println("Execute REST query DONE");
            return response.getStatusLine().getStatusCode();
        }
        catch (Throwable ex){
            Logger.getLogger(Exchange.class.getName()).log(Level.SEVERE, null, ex);
        }
        return 0;
    }

    @Override
    public void pushMessage(ExchangeMessage message) {
        this.transmitMsgs.add(message);
    }

    @Override
    public ExchangeMessage fetchMessage() {
        try{
            return this.recievedMsgs.take();
        }
        catch(InterruptedException ex){
            return null;
        }
    }

    private class CommunicationThread implements Runnable{

        @Override
        public void run() {
            Instant time = Instant.now();
            System.out.println("["+time+"]: Thread RUN");
            if(transmitMsgs.size()>0){
                System.out.println("["+time+"]: Started sending");
                postAllMessagesToEndpoint(MSGS_ENDPOINT);
                System.out.println("["+time+"]: Done sending");
            }
            else{
                System.out.println("["+time+"]: Nothing to send");
            }

            List<ExchangeMessage> msg = getAllMessagesFromEndpoint(MSGS_ENDPOINT);
            if(msg.size()>0){
                System.out.println("["+time+"]: Received messages: "+msg.size());
                recievedMsgs.addAll(msg);
            }
            System.out.println("["+time+"]: Thread DONE");
        }
    }

    private class ExchangeThreadFactory implements ThreadFactory{

        @Override
        public Thread newThread(Runnable r) {
            Thread T = new Thread(r);
            T.setDaemon(true);
            return T;
        }

    }
}
